"""
Taken from dirichlet python package for local use.
Huge thanks to the developer.

"""


"""Dirichlet.py

Maximum likelihood estimation and likelihood ratio tests of Dirichlet
distribution models of data.

Most of this package is a port of Thomas P. Minka's wonderful Fastfit MATLAB
code. Much thanks to him for that and his clear paper "Estimating a Dirichlet
distribution". See the following URL for more information:

    http://research.microsoft.com/en-us/um/people/minka/"""

import sys

import numpy as np
import scipy as sp
import scipy.stats as stats
from numpy import (
    arange,
    array,
    asanyarray,
    asarray,
    diag,
    exp,
    isscalar,
    log,
    ndarray,
    ones,
    vstack,
    zeros,
)
from numpy.linalg import norm
from scipy.special import gammaln, polygamma, psi

from . import simplex
import warnings

MAXINT = sys.maxsize

__all__ = [
    "loglikelihood",
    "meanprecision",
    "mle",
    "pdf",
    "test",
]

euler = -1 * psi(1)  # Euler-Mascheroni constant


class NotConvergingError(Exception):
    """Error when a successive approximation method doesn't converge
    """
    pass


def test(D1, D2, method="meanprecision", maxiter=None):
    """Test for statistical difference between observed proportions.

    Parameters
    ----------
    D1 : (N1, K) shape array
    D2 : (N2, K) shape array
        Input observations. ``N1`` and ``N2`` are the number of observations,
        and ``K`` is the number of parameters for the Dirichlet distribution
        (i.e. the number of levels or categorical possibilities).
        Each cell is the proportion seen in that category for a particular
        observation. Rows of the matrices must add up to 1.
    method : string
        One of ``'fixedpoint'`` and ``'meanprecision'``, designates method by
        which to find MLE Dirichlet distribution. Default is
        ``'meanprecision'``, which is faster.
    maxiter : int
        Maximum number of iterations to take calculations. Default is
        ``sys.maxint``.

    Returns
    -------
    D : float
        Test statistic, which is ``-2 * log`` of likelihood ratios.
    p : float
        p-value of test.
    a0 : (K,) shape array
    a1 : (K,) shape array
    a2 : (K,) shape array
        MLE parameters for the Dirichlet distributions fit to 
        ``D1`` and ``D2`` together, ``D1``, and ``D2``, respectively."""

    N1, K1 = D1.shape
    N2, K2 = D2.shape
    if K1 != K2:
        raise ValueError("D1 and D2 must have the same number of columns")

    D0 = vstack((D1, D2))
    a0 = mle(D0, method=method, maxiter=maxiter)
    a1 = mle(D1, method=method, maxiter=maxiter)
    a2 = mle(D2, method=method, maxiter=maxiter)

    D = 2 * (loglikelihood(D1, a1) + loglikelihood(D2, a2) - loglikelihood(D0, a0))
    return (D, stats.chi2.sf(D, K1), a0, a1, a2)


def pdf(alphas):
    """Returns a Dirichlet PDF function

    Parameters
    ----------
    alphas : (K,) shape array
        The parameters for the distribution of shape ``(K,)``.

    Returns
    -------
    function
        The PDF function, takes an ``(N, K)`` shape input and gives an
        ``(N,)`` output.
    """
    alphap = alphas - 1
    c = np.exp(gammaln(alphas.sum()) - gammaln(alphas).sum())

    def dirichlet(xs):
        """Dirichlet PDF

        Parameters
        ----------
        xs : (N, K) shape array
            The ``(N, K)`` shape input matrix
        
        Returns
        -------
        (N,) shape array
            Point value for PDF
        """
        return c * (xs ** alphap).prod(axis=1)

    return dirichlet


def meanprecision(a):
    """Mean and precision of a Dirichlet distribution.

    Parameters
    ----------
    a : (K,) shape array
        Parameters of a Dirichlet distribution.

    Returns
    -------
    mean : (K,) shape array
        Means of the Dirichlet distribution. Values are in [0,1].
    precision : float
        Precision or concentration parameter of the Dirichlet distribution."""

    s = a.sum()
    m = a / s
    return (m, s)


def loglikelihood(D, a):
    """Compute log likelihood of Dirichlet distribution, i.e. log p(D|a).

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    a : (K,) shape array
        Parameters for the Dirichlet distribution.

    Returns
    -------
    logl : float
        The log likelihood of the Dirichlet distribution"""
    N, K = D.shape
    logp = log(D).mean(axis=0)
    return N * (gammaln(a.sum()) - gammaln(a).sum() + ((a - 1) * logp).sum())


def mle(D, tol=1e-7, method="meanprecision", maxiter=None):
    """Iteratively computes maximum likelihood Dirichlet distribution
    for an observed data set, i.e. a for which log p(D|a) is maximum.

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    method : string
        One of ``'fixedpoint'`` and ``'meanprecision'``, designates method by
        which to find MLE Dirichlet distribution. Default is
        ``'meanprecision'``, which is faster.
    maxiter : int
        Maximum number of iterations to take calculations. Default is
        ``sys.maxint``.

    Returns
    -------
    a : (K,) shape array
        Maximum likelihood parameters for Dirichlet distribution."""

    if method == "meanprecision":
        return _meanprecision(D, tol=tol, maxiter=maxiter)
    else:
        return _fixedpoint(D, tol=tol, maxiter=maxiter)


def _fixedpoint(D, tol=1e-7, maxiter=None):
    """Simple fixed point iteration method for MLE of Dirichlet distribution

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is
        ``sys.maxint``.

    Returns
    -------
    a : (K,) shape array
        Fixed-point estimated parameters for Dirichlet distribution."""
    logp = log(D).mean(axis=0)
    a0 = _init_a(D)

    # Start updating
    if maxiter is None:
        maxiter = MAXINT
    for i in range(maxiter):
        a1 = _ipsi(psi(a0.sum()) + logp)
        # Much faster convergence than with the more obvious condition
        # `norm(a1-a0) < tol`
        if abs(loglikelihood(D, a1) - loglikelihood(D, a0)) < tol:
            return a1
        a0 = a1
    warnings.warn(
        "Failed to converge after {} iterations, values are {}.".format(maxiter, a1)
    )
    return a1


def _meanprecision(D, tol=1e-7, maxiter=None):
    """Mean/precision method for MLE of Dirichlet distribution

    Uses alternating estimations of mean and precision.

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is
        ``sys.maxint``.

    Returns
    -------
    a : (K,) shape array
        Estimated parameters for Dirichlet distribution."""
    logp = log(D).mean(axis=0)
    a0 = _init_a(D)
    s0 = a0.sum()
    if s0 < 0:
        a0 = a0 / s0
        s0 = 1
    elif s0 == 0:
        a0 = ones(a0.shape) / len(a0)
        s0 = 1
    m0 = a0 / s0

    # Start updating
    if maxiter is None:
        maxiter = MAXINT
    for i in range(maxiter):
        a1 = _fit_s(D, a0, logp, tol=tol)
        s1 = sum(a1)
        a1 = _fit_m(D, a1, logp, tol=tol)
        m = a1 / s1
        # Much faster convergence than with the more obvious condition
        # `norm(a1-a0) < tol`
        if abs(loglikelihood(D, a1) - loglikelihood(D, a0)) < tol:
            return a1
        a0 = a1
    warnings.warn(
        f"Failed to converge after {maxiter} iterations, " f"values are {a1}."
    )
    return a1


def _fit_s(D, a0, logp, tol=1e-7, maxiter=1000):
    """Update parameters via MLE of precision with fixed mean

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    a0 : (K,) shape array
        Current parameters for Dirichlet distribution
    logp : (K,) shape array
        Mean of log-transformed D across N observations
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is 1000.

    Returns
    -------
    (K,) shape array
        Updated parameters for Dirichlet distribution."""
    s1 = a0.sum()
    m = a0 / s1
    mlogp = (m * logp).sum()
    for i in range(maxiter):
        s0 = s1
        g = psi(s1) - (m * psi(s1 * m)).sum() + mlogp
        h = _trigamma(s1) - ((m ** 2) * _trigamma(s1 * m)).sum()

        if g + s1 * h < 0:
            s1 = 1 / (1 / s0 + g / h / (s0 ** 2))
        if s1 <= 0:
            s1 = s0 * exp(-g / (s0 * h + g))  # Newton on log s
        if s1 <= 0:
            s1 = 1 / (1 / s0 + g / ((s0 ** 2) * h + 2 * s0 * g))  # Newton on 1/s
        if s1 <= 0:
            s1 = s0 - g / h  # Newton
        if s1 <= 0:
            raise NotConvergingError(f"Unable to update s from {s0}")

        a = s1 * m
        if abs(s1 - s0) < tol:
            return a
    warnings.warn(f"Failed to converge after {maxiter} iterations, " f"s is {s1}")
    return a


def _fit_m(D, a0, logp, tol=1e-7, maxiter=1000):
    """Update parameters via MLE of mean with fixed precision s

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    a0 : (K,) shape array
        Current parameters for Dirichlet distribution
    logp : (K,) shape array
        Mean of log-transformed D across N observations
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is 1000.

    Returns
    -------
    (K,) shape array
        Updated parameters for Dirichlet distribution."""
    s = a0.sum()
    for i in range(maxiter):
        m = a0 / s
        a1 = _ipsi(logp + (m * (psi(a0) - logp)).sum())
        a1 = a1 / a1.sum() * s

        if norm(a1 - a0) < tol:
            return a1
        a0 = a1
    warnings.warn(f"Failed to converge after {maxiter} iterations, " f"s is {s}")
    return a1


def _init_a(D):
    """Initial guess for Dirichlet alpha parameters given data D

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.

    Returns
    -------
    (K,) shape array
        Crude guess for parameters of Dirichlet distribution."""
    E = D.mean(axis=0)
    E2 = (D ** 2).mean(axis=0)
    return ((E[0] - E2[0]) / (E2[0] - E[0] ** 2)) * E


def _ipsi(y, tol=1.48e-9, maxiter=10):
    """Inverse of psi (digamma) using Newton's method. For the purposes
    of Dirichlet MLE, since the parameters a[i] must always
    satisfy a > 0, we define ipsi :: R -> (0,inf).

    Parameters
    ----------
    y : (K,) shape array
        y-values of psi(x)
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is 10.

    Returns
    -------
    (K,) shape array
        Approximate x for psi(x)."""
    y = asanyarray(y, dtype="float")
    x0 = np.piecewise(
        y,
        [y >= -2.22, y < -2.22],
        [(lambda x: exp(x) + 0.5), (lambda x: -1 / (x + euler))],
    )
    for i in range(maxiter):
        x1 = x0 - (psi(x0) - y) / _trigamma(x0)
        if norm(x1 - x0) < tol:
            return x1
        x0 = x1
    warnings.warn(
        f"Failed to converge after {maxiter} iterations, " f"value is {x1}")
    return x1


def _trigamma(x):
    return polygamma(1, x)



"""
Most of the codes are used from git repository mvem. Huge thanks to the
devoloper. 

"""

import numpy as np
import scipy.stats
from scipy.special import gammaln, kv, digamma
from fmvmm.distributions.py_gig import rgig


def custom_log(input_value):
    # if np.any(input_value,where=0):
    return np.log1p(input_value)
    # else:
    #     return np.log(input_value)

def _gig_to_scipy_params_(lmbda, chi, psi):
    """
    Transform from (chi, psi) parameterisations to (loc, scale)
    """
    p = lmbda
    b = np.sqrt(chi*psi)
    #loc = 0
    scale = np.sqrt(chi/psi)
    #alpha = np.sqrt(chi/psi)
    beta2 = psi*chi
    beta =np.sqrt(psi*chi);
    lm1 = lmbda - 1.0;
    lm12 = lm1*lm1;
    m = (lm1 + np.sqrt(lm12 + beta2))/beta;
    return p, b, m, scale

def _check_gig_pars(lmbda, chi, psi):
    """
    Check parameter consistency.
    """
    if (lmbda < 0) and (chi <= 0 or psi < 0):
        raise Exception("If lambda < 0: chi must be > 0 and psi must be >= 0! \n" + \
                        "lambda = " + str(lmbda) + "; chi = " + str(chi) + "; psi = " + str(psi) + "\n")
    if (lmbda == 0) and (chi <= 0 or psi <= 0):
        raise Exception("If lambda == 0: chi must be > 0 and psi must be > 0! \n" + \
                        "lambda = " + str(lmbda) + "; chi = " + str(chi) + "; psi = " + str(psi) + "\n")
    if (lmbda > 0) and (chi < 0 or psi <= 0):
        raise Exception("If lambda > 0: chi must be >= 0 and psi must be > 0! \n" + \
                        "lambda = " + str(lmbda) + "; chi = " + str(chi) + "; psi = " + str(psi) + "\n")

def logpdf(x, lmbda=1, chi=1, psi=1):
    """
    Log-probability density function of the GIG distribution.

    :param x: An array of shape (n,) containing n observations of some
        univariate data.
    :type x: np.ndarray
    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :return: The log-density at each observation.
    :rtype: np.ndarray with shape (n,).
    """
    _check_gig_pars(lmbda, chi, psi)

    if psi==0:
        # inverse gamma (students t case)
        logdensity = scipy.stats.invgamma.logpdf(x, a=-lmbda, scale=0.5*chi)
    elif chi==0:
        # gamma (VG case)
        logdensity = scipy.stats.gamma.logpdf(x, a=lmbda, scale=psi/2)
    else:
        p, b, loc, scale = _gig_to_scipy_params_(lmbda, chi, psi)
        logdensity = scipy.stats.geninvgauss.logpdf(x, p=p, b=b, loc=loc, scale=scale)

    return logdensity

# dgig
def pdf(x, lmbda=1, chi=1, psi=1):
    """
    Probability density function of the GIG distribution.

    :param x: An array of shape (n,) containing n observations of some
        univariate data.
    :type x: np.ndarray
    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :return: The density at each observation.
    :rtype: np.ndarray with shape (n,).
    """

    logdensity = logpdf(x, lmbda, chi, psi)
    return np.exp(logdensity)

def logcdf(x, lmbda=1, chi=1, psi=1):
    """
    Log-cumulative density function of the GIG distribution.

    :param x: An array of shape (n,) containing n observations of some
        univariate data.
    :type x: np.ndarray
    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :return: The log-cumulative density at each observation.
    :rtype: np.ndarray with shape (n,).
    """

    _check_gig_pars(lmbda, chi, psi)

    if psi==0:
        # inverse gamma (students t case)
        return scipy.stats.invgamma.logcdf(x, a=-lmbda, scale=0.5*chi)
    elif chi==0:
        # gamma (VG case)
        return scipy.stats.gamma.logcdf(x, a=lmbda, scale=psi/2)
    else:
        p, b, loc, scale = _gig_to_scipy_params_(lmbda, chi, psi)
        return scipy.stats.geninvgauss.logcdf(x, p=p, b=b, loc=loc, scale=scale)

# pgig
def cdf(x, lmbda=1, chi=1, psi=1):
    """
    Cumulative density function of the GIG distribution.

    :param x: An array of shape (n,) containing n observations of some
        univariate data.
    :type x: np.ndarray
    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :return: The cumulative density at each observation.
    :rtype: np.ndarray with shape (n,).
    """

    lgcdf = logcdf(x, lmbda, chi, psi)
    return np.exp(lgcdf)

def rvs(lmbda=1, chi=1, psi=1, size=1):
    """
    Random number generator of the GIG distribution.

    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :param size: The number of samples to draw. Defaults to 1.
    :type size: int, optional
    :return: The random univariate numbers generated.
    :rtype: np.ndarray with shape (n,).
    """

    _check_gig_pars(lmbda, chi, psi)

    if psi==0:
        # inverse gamma (students t case)
        return scipy.stats.gamma.rvs(a=-lmbda, scale=chi/2, size=size)
    elif chi==0:
        # gamma (VG case)
        return scipy.stats.gamma.rvs(a=lmbda, scale=psi/2, size=size)
    else:
        p, b, loc, scale = _gig_to_scipy_params_(lmbda, chi, psi)
        return scipy.stats.geninvgauss.rvs(p=p, b=b, loc=loc, scale=scale, size=size)
        #return rgig(size,lmbda,chi,psi)

def _besselM3(lmbda=9/2, x=2, logvalue=False):
    if np.all(np.abs(lmbda)==0.5):
        ## Simplified expression in case lambda == 0.5
        if not logvalue:
            res = np.sqrt(np.pi/(2*x))*np.exp(-x)
        else:
            res = 0.5*np.log(np.pi/(2*x))-x
    else:
        if not logvalue:
            res = kv(lmbda, x)
        else:
            res = np.log(kv(lmbda, x))
            # res = custom_log(kv(lmbda, x))
    return res

def expect(lmbda, chi, psi, func = "x"):
    """
    Compute the expectation E[f(x)|lmbda, chi, psi],
    where f(x) is one of ["x", "logx", "1/x"].

    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :param func: The function to compute the expectation of. One
        of ["x", "logx", "1/x"]. Default to "x".
    :type func: str, optional
    :return: The expected value
    :rtype: float
    """

    if func not in ["x", "logx", "1/x"]:
        raise Exception("_func_ must be one of ['x', 'logx', '1/x'], but is: " + str(func))

    lmbda = np.asarray(lmbda)
    chi = np.asarray(chi)
    psi = np.asarray(psi)

    #for i in range(len(lmbda)):
    #    _check_gig_pars(lmbda[i], chi[i], psi[i])

    chi_eps = 1e-8


    if func == "x":
        if np.all(psi == 0):
            # inv Gamma (Student-t)
            beta = 0.5 * chi
            alpha = -lmbda
            if alpha <= 1:
                alpha = np.nan # expectation is not defined if alpha <= 1
            return beta/(alpha-1)

        elif np.all(chi == 0):
            # gamma (VG)
            beta = 0.5 * psi
            alpha = lmbda
            return alpha/beta

        else:
            # gig (ghyp, hyp, NIG)
            chi[np.abs(chi) < chi_eps] = np.sign(chi[np.abs(chi) < chi_eps]) * chi_eps
            chi[chi == 0] = chi_eps

            alpha_bar = np.sqrt(chi * psi)
            term1 = 0.5 * np.log(chi / psi)
            term2 = _besselM3(lmbda + 1, alpha_bar, logvalue = True)
            term3 = _besselM3(lmbda, alpha_bar, logvalue = True)

            return np.exp(term1 + term2 - term3)

    elif func == "logx":
        if np.all(psi == 0):
            ## Inv Gamma -> Student-t
            beta = 0.5 * chi
            alpha = -lmbda
            return np.log(beta) - digamma(alpha)
        elif np.all(chi == 0):
            ## Gamma -> VG
            beta = 0.5 * psi
            alpha = lmbda
            return digamma(alpha) - np.log(beta)
        else:
            ## GIG -> ghyp, hyp, NIG
            
            # print("# doing GIG -> ghyp, hyp, NIG")
            
            chi[np.abs(chi) < chi_eps] = np.sign(chi[np.abs(chi) < chi_eps]) * chi_eps
            chi[chi == 0] = chi_eps

            # requires numerical calculations of the derivative of E[X**a] for a=sqrt(chi*psi)
            alpha_bar = np.sqrt(chi * psi)
            besselKnu = lambda x, alpha_bar: kv(x, alpha_bar)
            #Kderiv = np.concatenate([scipy.optimize.approx_fprime(lmbda, besselKnu, 1e-8, a) for a in alpha_bar])
            step = 1e-8
            Kderiv = (besselKnu(lmbda+step, alpha_bar) - besselKnu(lmbda-step, alpha_bar)) / (2*step)
            return 0.5 * np.log(chi/psi) + Kderiv / kv(lmbda, alpha_bar)

    elif func == "1/x":
        if np.all(psi == 0):
            # Inv Gamma -> Student-t
            beta = 0.5 * chi
            alpha = -lmbda
            return alpha / beta
        elif np.all(chi == 0):
            ## Gamma -> VG
            print("Case 'chi == 0' and 'func = 1/x' is not implemented, using scipy function...")
            return scipy.stats.gamma.expect(lambda x: 1/x, args=(lmbda,), scale=psi/2)
        else:
            # GIG -> ghyp, hyp, NIG
            # print("# doing GIG -> ghyp, hyp, NIG")
            chi[np.abs(chi) < chi_eps] = np.sign(chi[np.abs(chi) < chi_eps]) * chi_eps
            chi[chi==0] = chi_eps

            alpha_bar = np.sqrt(chi * psi)
            # print("abar",np.mean(alpha_bar))
            term1 = -0.5 * np.log(chi / psi)
            term2 = _besselM3(lmbda - 1, alpha_bar, logvalue = True)
            # print("term2",np.mean(term2))
            term3 = _besselM3(lmbda, alpha_bar, logvalue = True)
            # print("term3",np.mean(term3))
            return np.exp(term1 + term2 - term3)


def var(lmbda, chi, psi):
    """
    Compute the variance of a GIG distribution with parameters
    (lmbda, chi, psi). When psi==0, the distribution is the inverse
    Gamma distribution, and when chi==0, the distribution is the Gamma
    distribution.

    Handles a vector of parameters, if the different distributions all
    belong to the same group. That is, if chi==0 for one set of parameters,
    it must be so for all sets of parameters.

    :param lmbda: Univariate parameter.
    :type lmbda: float
    :param chi: Univariate parameter.
    :type chi: float
    :param psi: Univariate parameter.
    :type psi: float
    :return: The variance
    :rtype: float
    """

    lmbda = np.asarray(lmbda)
    chi = np.asarray(chi)
    psi = np.asarray(psi)

    for i in range(len(lmbda)):
        _check_gig_pars(lmbda[i], chi[i], psi[i])

    chi_eps = 1e-8

    if np.all(psi == 0):
        ## Inv Gamma -> Student-t
        beta = 0.5 * chi
        alpha = -lmbda

        tmp_var = beta**2 / ((alpha - 1)**2 * (alpha - 2))
        tmp_var[alpha <= 2] = np.inf # variance is Inf if alpha <= 2
        return tmp_var

    elif np.all(chi == 0):
        ## Gamma -> VG
        beta = 0.5 * psi
        alpha = lmbda
        return alpha/(beta**2)
    else:
        ## GIG -> ghyp, hyp, NIG
        chi[np.abs(chi) < chi_eps] = np.sign(chi[np.abs(chi) < chi_eps]) * chi_eps
        chi[chi==0] = chi_eps

        alpha_bar = np.sqrt(chi * psi)
        term1 = 0.5 * np.log(chi / psi)
        term2 = _besselM3(lmbda + 1, alpha_bar, logvalue = True)
        term3 = _besselM3(lmbda, alpha_bar, logvalue = True)
        var_term1 = np.log(chi / psi)
        var_term2 = _besselM3(lmbda + 2, alpha_bar, logvalue = True)
        var_term3 = _besselM3(lmbda, alpha_bar, logvalue = True)

        return np.exp(var_term1 + var_term2 - var_term3) - np.exp(term1 + term2 - term3)**2



from setuptools import setup, find_packages

# Read long description from README.md if available
with open("readme.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

project_url = {
    "Github": "https://github.com/samyajoypal/fmvmm",
}

setup(
    name="fmvmm",
    version="2.0.0",
    author="Samyajoy Pal",
    author_email="palsamyajoy@gmail.com",
    description="Flexible Multivariate Mixture Model",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/samyajoypal/fmvmm",
    project_urls=project_url,
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "conorm~=1.2.0",
        "matplotlib~=3.7",
        "numpy>=1.24, <1.26",
        "pandas>=2.0, <2.2",
        "scikit-learn>=1.6, <1.8",
        "scipy>=1.9, <1.11",
        "setuptools",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "my_command = fmvmm.command:main",
        ],
    },
    include_package_data=True,
)



import numpy as np
from sklearn.cluster import KMeans
from sklearn import mixture
from scipy.stats import dirichlet
import math
import pandas as pd
import conorm
from scipy.special import gammaln, psi,digamma
import dirichlet as drm
import scipy.special as sp


def hard_assignments(gamma):
    """
    Convert soft responsibilities (rows sum to 1) into hard 0/1 assignments
    by taking argmax in each row. Returns a new array with exactly one '1'
    per row, and zeros elsewhere.
    """
    n_data, k = gamma.shape
    hard = np.zeros_like(gamma)
    max_indices = np.argmax(gamma, axis=1)
    hard[np.arange(n_data), max_indices] = 1.0
    return hard

def mixture_counts(gamma, mode='soft'):
    """
    Depending on mode='soft' or 'hard', return the cluster counts N_j for j=1..k
    and the total N = sum_j N_j.
    """
    if mode == 'hard':
        gamma_hard = hard_assignments(gamma)
        N_j = np.sum(gamma_hard, axis=0)  # integer counts
    else:
        N_j = np.sum(gamma, axis=0)      # sum of partial responsibilities
    N = np.sum(N_j)
    return N_j, N

def mixture_proportions_info(pi, N_j):
    """
    For mixture weights pi (length k), compute the unconstrained (k x k)
    information matrix I_pi and its inverse, using:
       I_pi       = diag(N_j / pi_j^2),
       I_pi_inv   = diag(pi_j^2 / N_j).
    """
    # pi, N_j should each be shape (k,)
    I_diag = N_j / (pi**2)
    I_pi = np.diag(I_diag)
    
    I_inv_diag = (pi**2) / N_j
    I_pi_inv = np.diag(I_inv_diag)
    
    return I_pi, I_pi_inv

def single_dirichlet_info(alpha_j, n_j, ridge_factor=1e-10, use_pinv=False):
    """
    Compute the (d x d) Fisher info matrix and its inverse for ONE Dirichlet
    parameter vector alpha_j of length d, given 'n_j' as the effective sample size
    for that cluster (either sum of responsibilities if soft or integer if hard).

    Using the closed-form:
        I_alpha = D + G * 11^T,
        I_alpha^-1 = D_star + beta * (a_star)(a_star)^T,

    where:
      - D = diag(n_j * psi'(alpha_j)),
      - G = - n_j * psi'(sum(alpha_j)),
      - D_star = diag(1 / [n_j * psi'(alpha_j)] ),
      - a_star = [1 / psi'(alpha_j1), ..., 1 / psi'(alpha_jd]]^T,
      - beta = [n_j * psi'(sum alpha_j)] / [1 - psi'(sum alpha_j)* sum(1/psi'(alpha_j))].
    """
    d = len(alpha_j)
    alpha_sum = np.sum(alpha_j)
    
    # D = diag(n_j * trigamma(alpha_j))
    D_vals = n_j * sp.polygamma(1, alpha_j)  # shape (d,)
    D = np.diag(D_vals)
    
    # G = - n_j * trigamma(sum_j alpha_j)
    psi_sum = sp.polygamma(1, alpha_sum)
    G = - n_j * psi_sum
    
    # => I_alpha (d x d)
    I_alpha = D + G * np.ones((d, d))
    
    # # Inverse
    # D_star = np.diag(1.0 / D_vals)  # shape (d x d)
    # a_star_vals = 1.0 / sp.polygamma(1, alpha_j)  # shape (d,)
    # a_star = a_star_vals.reshape(-1,1)            # (d,1)

    # denom = 1.0 - psi_sum * np.sum(a_star_vals)
    # if abs(denom) < 1e-15:
    #     raise ValueError("Denominator for beta is near zero; check alpha_j range or n_j.")
    
    # beta = (n_j * psi_sum) / denom
    # I_alpha_inv = D_star + beta * (a_star @ a_star.T)
    
    # I_alpha_inv = np.linalg.inv(I_alpha)
    M = I_alpha.shape[0]
    # First attempt direct inverse
    try:
        I_alpha_inv = np.linalg.inv(I_alpha)
    except np.linalg.LinAlgError:
        # Direct inverse failed
        if use_pinv:
            # Use the pseudoinverse
            I_alpha_inv = np.linalg.pinv(I_alpha)
        else:
            # Try adding a small ridge on the diagonal
            ridge_mat = I_alpha + ridge_factor*np.eye(M)
            # In case that also fails, we do a nested try:
            try:
                I_alpha_inv = np.linalg.inv(ridge_mat)
            except np.linalg.LinAlgError:
                # As a last fallback, use pseudoinverse of the ridge version
                I_alpha_inv = np.linalg.pinv(ridge_mat)
    
    return I_alpha, I_alpha_inv

def combined_info_and_se(pi, alpha, gamma, mode='soft'):
    """
    Build and return a *combined* Fisher Info matrix for:
        - Mixture proportions pi (length k)
        - Dirichlet parameters alpha (shape (k, d)), i.e. alpha_j is the d-dim vector for cluster j

    We do this by forming a block-diagonal matrix of size [k + k*d,  k + k*d]:
      - Top-left (k x k): mixture_proportions_info
      - Then k blocks along the diagonal (d x d each) for each cluster's alpha_j
    Cross terms are assumed zero in a standard Dirichlet mixture model.

    Returns:
       I_total, I_total_inv, se_total

    where se_total is an array of length [k + k*d] giving the sqrt of the diagonal
    of I_total_inv. The ordering is:
        [ pi_1, ..., pi_k, alpha_1(1),...,alpha_1(d), alpha_2(1),..., alpha_k(d) ].
    """
    # 1) First, gather mixture counts
    N_j, N = mixture_counts(gamma, mode=mode)
    k = len(pi)         # number of mixture components
    # alpha should be shape (k, d)
    if alpha.ndim != 2 or alpha.shape[0] != k:
        raise ValueError("alpha must be shape (k, d). You gave shape %s" % (alpha.shape,))

    d = alpha.shape[1]  # dimension of each Dirichlet
    # We'll produce a block-diagonal of dimension k + k*d

    # 2) Mixture block
    I_pi, I_pi_inv = mixture_proportions_info(pi, N_j)

    # 3) Dirichlet blocks
    # For each cluster j, we compute a single_dirichlet_info with n_j = N_j[j].
    # Then place that block into the big matrix along the diagonal.
    I_blocks = []
    I_inv_blocks = []
    for j in range(k):
        alpha_j = alpha[j]          # shape (d,)
        n_j = N_j[j]                # either integer or fractional
        I_j, I_j_inv = single_dirichlet_info(alpha_j, n_j)
        I_blocks.append(I_j)
        I_inv_blocks.append(I_j_inv)

    # 4) Combine into a big block-diagonal
    big_dim = k + k*d
    I_total = np.zeros((big_dim, big_dim))
    I_total_inv = np.zeros((big_dim, big_dim))
    
    # (a) place the (k x k) mixture block at top-left
    I_total[0:k, 0:k] = I_pi
    I_total_inv[0:k, 0:k] = I_pi_inv

    # (b) place each (d x d) block for alpha_j
    #    index offset in the big matrix: row,col start = k + j*d
    for j in range(k):
        row_start = k + j*d
        row_end   = k + (j+1)*d
        I_total[row_start:row_end, row_start:row_end] = I_blocks[j]
        I_total_inv[row_start:row_end, row_start:row_end] = I_inv_blocks[j]

    # 5) Standard errors from diagonal of the inverse
    var_diag = np.diag(I_total_inv)
    se_total = np.sqrt(var_diag)

    return I_total, se_total

def dirichlet_kl_divergence(alpha1, alpha2):
    """
    Compute KL divergence between two Dirichlet distributions.

    Parameters:
        alpha1 (numpy array): Parameters of the first Dirichlet distribution.
        alpha2 (numpy array): Parameters of the second Dirichlet distribution.

    Returns:
        float: KL divergence between the two Dirichlet distributions.
    """
    sum_alpha1 = np.sum(alpha1)
    sum_alpha2 = np.sum(alpha2)

    term1 = gammaln(sum_alpha1) - gammaln(sum_alpha2)
    term2 = np.sum(gammaln(alpha2) - gammaln(alpha1))
    term3 = np.sum((alpha1 - alpha2) * (digamma(alpha1) - digamma(sum_alpha1)))

    kl_divergence = term1 + term2 + term3
    return kl_divergence

def dmm_kl_divergence(pi, omega, f_models, g_models):
    """
    Compute KL divergence between two Dirichlet mixture models.

    Parameters:
        pi (numpy array): Mixing coefficients for the first mixture model.
        omega (numpy array): Mixing coefficients for the second mixture model.
        f_models (list of numpy arrays): Parameters of Dirichlet distributions in the first mixture model.
        g_models (list of numpy arrays): Parameters of Dirichlet distributions in the second mixture model.

    Returns:
        float: KL divergence between the two Dirichlet mixture models.
    """
    kl_divergence = 0.0

    for a in range(len(pi)):
        term_num = 0.0
        term_denom = 0.0

        for aprime in range(len(pi)):
            term_num += pi[aprime] * np.exp(-dirichlet_kl_divergence(f_models[a], f_models[aprime]))

        for b in range(len(omega)):
            term_denom += omega[b] * np.exp(-dirichlet_kl_divergence(f_models[a], g_models[b]))
        
        if term_denom==0:
            term_denom=term_denom+1e-243  
        if term_num==0:
            term_num=term_num+1e-243
        kl_divergence += pi[a] * np.log(term_num / term_denom)

    return kl_divergence


def dmm_mc_kl_divergence(pi, omega, f_models, g_models, n_samples=1000):
    """
    Monte Carlo estimation of KL Divergence between two Dirichlet mixture models.

    Parameters:
        pi (numpy array): Mixing coefficients for the first mixture model.
        omega (numpy array): Mixing coefficients for the second mixture model.
        f_models (2D numpy array): Parameters of Dirichlet distributions in the first mixture model.
        g_models (2D numpy array): Parameters of Dirichlet distributions in the second mixture model.
        n_samples (int): Number of samples to draw.

    Returns:
        float: Estimated KL Divergence.
    """
    np.random.seed(0)
    k=len(pi)
    assert n_samples>k
    nis=[int(n_samples*i) for i in pi]
    xis=[]
    for j in range(k):
        xis.extend(np.random.dirichlet(f_models[j],nis[j]))
    
    logs = []

    for xi in xis:
        # # Draw a sample xi from the mixture model using pi and f_models
        # xi_=[pi[j] * np.random.dirichlet(f_models[j]) for j in range(len(pi))]
        # xi = np.sum(xi_,axis=0)
        # # print(xi_)
        # Calculate f_pdf and g_pdf
        with np.errstate(under='ignore'):
            f_pdf_=[pi[j] * dirichlet.pdf(xi, f_models[j]) for j in range(len(pi))]
            f_pdf = np.nansum(np.array(f_pdf_,dtype=np.float128))
            g_pdf_=[omega[j] * dirichlet.pdf(xi, g_models[j]) for j in range(len(omega))]
            g_pdf = np.nansum(np.array(g_pdf_,dtype=np.float128))
    
            # Calculate log(f(x_i)/g(x_i)) and append to logs
            logs.append(np.log(f_pdf / g_pdf))

    # Average over all samples
    return np.nansum(logs)/len(logs)



def sample_dirichlet_gibbs(alpha, n_iterations=1000):
  """
  This function draws samples from a Dirichlet distribution using the Gibbs sampler and averages them.

  Args:
      alpha: A numpy array of length n_components representing the Dirichlet parameters.
      n_samples: The number of samples to draw.

  Returns:
      A numpy array of shape (n_samples, n_components) containing the average of the samples.
  """

  # Initialize an empty list to store samples
  samples = []

  # Iterate for the desired number of samples
  for _ in range(n_iterations):
    # Generate a single sample
    sample = _sample_dirichlet_gibbs(alpha)  # Helper function for one sample
    samples.append(sample)

  # Convert the list of samples to a numpy array and return the average
  return np.mean(np.array(samples), axis=0)


def _sample_dirichlet_gibbs(alpha, n_samples=1):
  """
  This function draws samples from a Dirichlet distribution using the Gibbs sampler.

  Args:
      alpha: A numpy array of length n_components representing the Dirichlet parameters.
      n_samples: The number of samples to draw.

  Returns:
      A numpy array of shape (n_samples, n_components) containing the samples.
  """
  n_components=len(alpha)
  # Initialize the samples matrix with random values from a uniform distribution
  samples = np.random.rand(n_samples, n_components)

  # Normalize each row to sum to 1 (represents probability distribution)
  samples /= samples.sum(axis=1, keepdims=True)

  for _ in range(n_samples - 1):
    # Iterate over each component
    for k in range(n_components):
      # Calculate the sum of alphas excluding the current component
      other_alpha_sum = alpha.sum() - alpha[k]
      # Sample from a beta distribution using the updated alpha values
      samples[:, k] = np.random.beta(alpha[k] + samples[:, :k].sum(axis=1), other_alpha_sum + samples[:, k+1:].sum(axis=1))

  return samples.flatten()


def dmm_mcmc_kl_divergence(pi, omega, f_models, g_models, n_samples=1000):
    """
    Monte Carlo estimation of KL Divergence between two Dirichlet mixture models.

    Parameters:
        pi (numpy array): Mixing coefficients for the first mixture model.
        omega (numpy array): Mixing coefficients for the second mixture model.
        f_models (2D numpy array): Parameters of Dirichlet distributions in the first mixture model.
        g_models (2D numpy array): Parameters of Dirichlet distributions in the second mixture model.
        n_samples (int): Number of samples to draw.

    Returns:
        float: Estimated KL Divergence.
    """
    logs = []
    np.random.seed(0)

    for _ in range(n_samples):
        # Draw a sample xi from the mixture model using pi and f_models
        xi = np.sum([pi[j] * sample_dirichlet_gibbs(f_models[j]) for j in range(len(pi))],axis=0)
        # print(xi)
        # Calculate f_pdf and g_pdf
        f_pdf = np.sum([pi[j] * dirichlet.pdf(xi, f_models[j]) for j in range(len(pi))])
        g_pdf = np.sum([omega[j] * dirichlet.pdf(xi, g_models[j]) for j in range(len(omega))])

        # Calculate log(f(x_i)/g(x_i)) and append to logs
        logs.append(np.log(f_pdf / g_pdf))

    # Average over all samples
    return np.mean(logs)



def closure(d_mat):
    d_mat = np.atleast_2d(d_mat)
    if np.any(d_mat < 0):
        raise ValueError("Cannot have negative proportions")
    if d_mat.ndim > 2:
        raise ValueError("Input matrix can only have two dimensions or less")
    if np.all(d_mat == 0, axis=1).sum() > 0:
        raise ValueError("Input matrix cannot have rows with all zeros")
    d_mat = d_mat / d_mat.sum(axis=1, keepdims=True)
    return d_mat.squeeze()

def multiplicative_replacement(d_mat, delta=None):
    d_mat = closure(d_mat)
    z_mat = (d_mat == 0)

    num_feats = d_mat.shape[-1]
    tot = z_mat.sum(axis=-1, keepdims=True)

    if delta is None:
        delta = (1. / num_feats)**2

    zcnts = 1 - tot * delta
    if np.any(zcnts) < 0:
        raise ValueError('The multiplicative replacement created negative '
                         'proportions. Consider using a smaller `delta`.')
    d_mat = np.where(z_mat, delta, zcnts * d_mat)
    return d_mat.squeeze()


def clr(mat):
    r"""
    Performs centre log ratio transformation.

    This function transforms compositions from Aitchison geometry to
    the real space. The :math:`clr` transform is both an isometry and an
    isomorphism defined on the following spaces

    :math:`clr: S^D \rightarrow U`

    where :math:`U=
    \{x :\sum\limits_{i=1}^D x = 0 \; \forall x \in \mathbb{R}^D\}`

    It is defined for a composition :math:`x` as follows:

    .. math::
        clr(x) = \ln\left[\frac{x_1}{g_m(x)}, \ldots, \frac{x_D}{g_m(x)}\right]

    where :math:`g_m(x) = (\prod\limits_{i=1}^{D} x_i)^{1/D}` is the geometric
    mean of :math:`x`.

    Parameters
    ----------
    mat : array_like, float
       a matrix of proportions where
       rows = compositions and
       columns = components

    Returns
    -------
    numpy.ndarray
         clr transformed matrix

    Examples
    --------
    >>> import numpy as np
    >>> from skbio.stats.composition import clr
    >>> x = np.array([.1, .3, .4, .2])
    >>> clr(x)
    array([-0.79451346,  0.30409883,  0.5917809 , -0.10136628])

    """
    mat = closure(mat)
    lmat = np.log(mat)
    gm = lmat.mean(axis=-1, keepdims=True)
    return (lmat - gm).squeeze()










def kmeans_init(data, k):
    n = len(data)
    p = len(data.columns)
    kmeans = KMeans(n_clusters=k, random_state=1)
    kmeans.fit(data)
    mu_not = kmeans.cluster_centers_
    #alsum=60
    alsum = p*5
    alpha_not = mu_not*alsum
    data_lol = data.values.tolist()
    data_cwise = [[] for i in range(k)]
    for i in range(n):
        data_cwise[kmeans.labels_[i]].append(data_lol[i])
    pi_not = [len(data_cwise[m])/n for m in range(k)]
    return pi_not, alpha_not


def gmm_init(data, k):
    n = len(data)
    clf = mixture.GaussianMixture(n_components=k, covariance_type='full')
    clf.fit(data)
    mu_not = clf.means_
    alsum = 60
    alpha_not = mu_not*alsum
    data_lol = data.values.tolist()
    data_cwise = [[] for i in range(k)]
    for i in range(n):
        data_cwise[clf.predict(data)[i]].append(data_lol[i])
    pi_not = [len(data_cwise[m])/n for m in range(k)]
    return pi_not, alpha_not


def random_init(data, k, random_seed=0):
    np.random.seed(random_seed)
    p = len(data.columns)
    alpha_not = []
    for h in range(k):
        alpha_not_temp = np.random.uniform(0, 50, p)
        alpha_not.append(alpha_not_temp)
    pi_not = sum(np.random.dirichlet([0.5 for i in range(k)], 1).tolist(), [])

    return pi_not, alpha_not


def kmeans_init_adv(data, k):
    n = len(data)
    kmeans = KMeans(n_clusters=k, random_state=1)
    kmeans.fit(data)
    #alpha_not=mu_not*alsum
    data_lol = data.tolist()
    data_cwise = [[] for i in range(k)]
    for i in range(n):
        data_cwise[kmeans.labels_[i]].append(data_lol[i])
    pi_not = [len(data_cwise[m])/n for m in range(k)]
    alpha_new = []
    for i in range(k):
        data_cwise_ar = np.array(data_cwise[i])
        E = data_cwise_ar.mean(axis=0)
        E2 = (data_cwise_ar ** 2).mean(axis=0)
        E3 = ((E[0] - E2[0]) / (E2[0] - E[0] ** 2)) * E
        alpha_new.append(E3.tolist())
    return pi_not, alpha_new


def gmm_init_adv(data, k):
    n = len(data)
    clf = mixture.GaussianMixture(n_components=k, covariance_type='full')
    clf.fit(data)
    #alpha_not=mu_not*alsum
    data_lol = data.tolist()
    data_cwise = [[] for i in range(k)]
    for i in range(n):
        data_cwise[clf.predict(data)[i]].append(data_lol[i])
    pi_not = [len(data_cwise[m])/n for m in range(k)]
    alpha_new = []
    for i in range(k):
        data_cwise_ar = np.array(data_cwise[i])
        E = data_cwise_ar.mean(axis=0)
        E2 = (data_cwise_ar ** 2).mean(axis=0)
        E3 = ((E[0] - E2[0]) / (E2[0] - E[0] ** 2)) * E
        alpha_new.append(E3.tolist())
    return pi_not, alpha_new


def dmm_loglikelihood(pi_temp, alpha_temp, data_lol):
    n = len(data_lol)
    k = len(alpha_temp)
    log_likelihood_values_temp = []
    for c in range(n):
        try:
            log_likelihood_old_temp = math.log(np.nansum(
                [pi_temp[f]*dirichlet.pdf(data_lol[c], alpha_temp[f]) for f in range(k)]))
        except:
            log_likelihood_old_temp = math.log1p(np.nansum(
                [pi_temp[f]*dirichlet.pdf(data_lol[c], alpha_temp[f]) for f in range(k)]))
        log_likelihood_values_temp.append(log_likelihood_old_temp)
    log_likelihood_old = np.sum(log_likelihood_values_temp)

    return log_likelihood_old


def dmm_responsibilities(pi_temp, alpha_temp, data_lol):
    n = len(data_lol)
    k = len(alpha_temp)
    gamma_temp = []
    for i in range(n):
        gamma_numer = []
        for j in range(k):
            temp_gamma_numer = (
                pi_temp[j]*dirichlet.pdf(data_lol[i], alpha_temp[j]))
            gamma_numer.append(temp_gamma_numer)
        gamma_row = gamma_numer / np.nansum(gamma_numer)
        gamma_temp.append(gamma_row)
    gamma_temp_ar = np.array(gamma_temp, dtype=np.float64)
    gamma_matrix = []
    for v in gamma_temp:
        gm_temp = v.tolist()
        gamma_matrix.append(gm_temp)
    return gamma_temp_ar, gamma_matrix


def dmm_pi_estimate(gamma_temp_ar):
    n = gamma_temp_ar.shape[0]
    k = gamma_temp_ar.shape[1]
    pi_new = []
    nk = []
    for g in range(k):
        nk_temp = np.nansum([gamma_temp_ar[w, g] for w in range(n)])
        pi_temp = nk_temp/n
        pi_new.append(pi_temp)
        nk.append(nk_temp)
    return pi_new


def count_to_comp(df):
    df_array=np.array(df)


    nf  = conorm.tmm_norm_factors(df)["norm.factors"]

    lj=[]
    for j in range(df_array.shape[1]):
        lj_temp=nf[j]*np.sum(df_array[:, j])
        lj.append(lj_temp)
        

    sj=[]
    for j in range(df_array.shape[1]):
        sj_temp=lj[j]/(np.sum(lj)/df_array.shape[1])
        sj.append(sj_temp)

    x_lol=[]
    for i in range(df_array.shape[0]):
        xi=[]
        for j in range(df_array.shape[1]):
            xi_temp=df_array[i,j]/sj[j]
            xi.append(xi_temp)
        xi_sum=np.sum(xi)
        xi_trans=[xi[k]/xi_sum for k in range(df_array.shape[1])]
       
            
        x_lol.append(xi_trans)
        #x_lol.append(xi)
        
    data=pd.DataFrame(x_lol)
    # trans_data=pd.DataFrame(multiplicative_replacement(data))
    trans_data=pd.DataFrame(data)
    
    return trans_data




def dirichlet_covariance(alpha):
    p = len(alpha)
    alpha0 = np.sum(alpha)
    cov = np.zeros((p, p))

    for i in range(p):
        for j in range(p):
            if i == j:
                cov[i, j] = (alpha[i] * (alpha0 - alpha[i])) / (alpha0**2 * (alpha0 + 1))
            else:
                cov[i, j] = -(alpha[i] * alpha[j]) / (alpha0**2 * (alpha0 + 1))

    return cov



import numpy as np
import pandas as pd
import itertools
import math
from sklearn.cluster import KMeans
from sklearn import mixture
from fmvmm.mixtures.mixmgh import approx_hessian_scipy
    
    


def fmm_loglikelihood(pi_temp,alpha_temp,data_lol,dist_comb):
    n = len(data_lol)
    k = len(alpha_temp)
    log_likelihood_values_temp = []
    for c in range(n):
        try:
            log_likelihood_old_temp = math.log(np.nansum(
                [pi_temp[f]*dist_comb[f].pdf(np.reshape(np.array(data_lol[c]),(1,len(data_lol[c]))), *alpha_temp[f]) for f in range(k)]))
        except:
            log_likelihood_old_temp = math.log1p(np.nansum(
                [pi_temp[f]*dist_comb[f].pdf(np.reshape(np.array(data_lol[c]),(1,len(data_lol[c]))), *alpha_temp[f]) for f in range(k)]))
        log_likelihood_values_temp.append(log_likelihood_old_temp)
    log_likelihood_old = np.sum(log_likelihood_values_temp)

    return log_likelihood_old


def fmm_responsibilities(pi_temp, alpha_temp, data_lol,dist_comb):
    n = len(data_lol)
    k = len(alpha_temp)
    gamma_temp = []
    for i in range(n):
        gamma_numer = []
        for j in range(k):
            temp_gamma_numer = (
                pi_temp[j]*dist_comb[j].pdf(np.reshape(np.array(data_lol[i]),(1,len(data_lol[i]))), *alpha_temp[j]))
            gamma_numer.append(temp_gamma_numer)
        gamma_row = gamma_numer / np.nansum(np.asanyarray(gamma_numer))
        gamma_temp.append(gamma_row)
    gamma_temp_ar = np.array(gamma_temp, dtype=np.float64)
    gamma_matrix = []
    for v in gamma_temp:
        gm_temp = v.tolist()
        gamma_matrix.append(gm_temp)
    return gamma_temp_ar, gamma_matrix


def fmm_pi_estimate(gamma_temp_ar,data_cwise):
    n = gamma_temp_ar.shape[0]
    k = gamma_temp_ar.shape[1]
    pi_new = []
    # nk = []
    # for g in range(k):
    #     nk_temp = np.nansum([gamma_temp_ar[w, g] for w in range(n)])
    #     pi_temp = nk_temp/n
    #     pi_new.append(pi_temp)
    #     nk.append(nk_temp)
    pi_new = [len(data_cwise[m])/n for m in range(k)]
    return pi_new


def fmm_kmeans_init(data, k,dist_comb):
    n = len(data)
    kmeans = KMeans(n_clusters=k, random_state=1)
    kmeans.fit(data)
    data_lol = data.tolist()
    data_cwise = [[] for i in range(k)]
    for i in range(n):
        data_cwise[kmeans.labels_[i]].append(data_lol[i])
    pi_not = [len(data_cwise[m])/n for m in range(k)]
    alpha_not=[dist_comb[j].fit(np.array(data_cwise[j])) for j in range(k)]
    return pi_not, alpha_not

def fmm_gmm_init(data, k,dist_comb):
    n = len(data)
    clf = mixture.GaussianMixture(n_components=k, covariance_type='full')
    clf.fit(data)
    data_lol = data.tolist()
    data_cwise = [[] for i in range(k)]
    for i in range(n):
        data_cwise[clf.predict(data)[i]].append(data_lol[i])
    pi_not = [len(data_cwise[m])/n for m in range(k)]
    alpha_not=[dist_comb[j].fit(np.array(data_cwise[j])) for j in range(k)]
    return pi_not, alpha_not



def fmm_estimate_alphas(data_cwise, alpha_not, dist_comb_not):
    alpha_new = []
    for t in range(len(data_cwise)):
        if np.array(data_cwise[t]).ndim == 1:
            alpha_new_temp = alpha_not[t]
        else:
            alpha_new_temp = dist_comb_not[t].fit(np.array(np.array(data_cwise[t])))

        alpha_new.append(alpha_new_temp)
    return alpha_new

def fmm_aic(alpha_new,log_likelihood_new, dist_comb):
    k = len(dist_comb)
    comp_params = np.sum([dist_comb[a].total_params(*alpha_new[a]) for a in range(len(dist_comb))])
    total_params = k-1 +  comp_params
    
    return -2 * log_likelihood_new + 2 * total_params
    



def fmm_bic(alpha_new,log_likelihood_new, dist_comb,n):
    k = len(dist_comb)
    comp_params = np.sum([dist_comb[a].total_params(*alpha_new[a]) for a in range(len(dist_comb))])
    total_params = k-1 +  comp_params

    return -2 * log_likelihood_new + total_params * np.log(n)

def fmm_icl(alpha_new,log_likelihood_new, dist_comb,n, gamma_temp_ar):
    
    bic_value = fmm_bic(alpha_new,log_likelihood_new, dist_comb,n)
    entropy_term = np.sum(gamma_temp_ar * np.log(np.clip(gamma_temp_ar, 1e-10, 1)))
    
    return bic_value + entropy_term
    
def full_neg_loglike(theta, data, param_shapes, lg_fun):
    n = data.shape[0]
    alpha_list = reshape_params(theta, param_shapes)
    ll = lg_fun(data, *alpha_list)
    total_ll = - np.sum(ll)
    
    return total_ll

def compute_info_scipy_fmvmm(logpdf_func, params_init, X, epsilon=1e-6):
    """
    Use SciPy to approximate the Hessian of the negative log-likelihood 
    for the GH mixture, then invert to get the empirical info matrix.

    Returns: I_e, Cov, SE
    """
    # Flatten final parameters
    n, p = X.shape
    param_shapes = [p.shape for p in params_init]
    theta_hat = flatten_params(params_init)

    d = len(theta_hat)

    # define partial function
    def nll_wrapper(th):
        return full_neg_loglike(th, X, param_shapes, logpdf_func)

    # compute Hessian
    H = approx_hessian_scipy(nll_wrapper, theta_hat, epsilon)

    return H

def flatten_params(params):
    """
    Flatten parameters into a single vector.
    
    Parameters:
        params (tuple): The parameters.
        
    Returns:
        flat_params (ndarray): The flattened parameters.
    """
    return np.array(np.concatenate([p.flatten() for p in params]),dtype=float)

def reshape_params(flat_params, param_shapes):
    """
    Reshape a flattened parameter vector into original shapes.
    
    Parameters:
        flat_params (ndarray): The flattened parameters.
        param_shapes (list): The shapes of the original parameters.
        
    Returns:
        params (tuple): The reshaped parameters.
    """
    params = []
    start = 0
    for shape in param_shapes:
        end = start + np.prod(shape)
        params.append(np.reshape(flat_params[start:end], shape))
        start = end
    return tuple(params)



from fmvmm.mixtures._base import BaseMixture
from fmvmm.mixtures.MixGHD import mainMGHD, MAPGH, llikGH
import time
import numpy as np
from fmvmm.distributions.multivariate_genhyperbolic import _alphabar2chipsi
from fmvmm.utils.utils_mixture import (mixture_clusters)
from fmvmm.distributions.multivariate_genhyperbolic import logpdf as logpdfmgh
from scipy.optimize import approx_fprime

def full_neg_loglike(theta, data, K, p):
    """
    Negative log-likelihood of the entire GH mixture model, 
    for flattened parameter vector theta.

    data: shape (n, p)
    K: number of mixture components
    p: dimension
    """
    pi_list, alpha_list = unflatten_mgh_params(theta, p, K)
    n = data.shape[0]

    total_ll = 0.0
    for i in range(n):
        x_i = data[i]  # shape (p,)
        mixture_val = 0.0
        for k in range(K):
            pi_k = pi_list[k]
            lmbd, chi, psi, mu, Sigma, gvec = alpha_list[k]
            # GH pdf => exp(logpdf)
            pdf_k = np.exp(logpdfmgh(x_i[None], lmbd, chi, psi, mu, Sigma, gvec))[0]
            mixture_val += pi_k * pdf_k
        # add -log(...) to total
        total_ll += -np.log(max(mixture_val, 1e-300))
    return total_ll

def approx_hessian_scipy(fun, x0, epsilon=1e-6):
    """
    Approximate Hessian of a scalar function fun(x) using nested calls to approx_fprime.
    fun: R^d -> scalar
    x0: shape(d,) initial point
    epsilon: step size for finite difference

    Returns: (d, d) array
    """
    d = len(x0)
    H = np.zeros((d, d), dtype=float)

    # baseline gradient at x0
    grad0 = approx_fprime(x0, fun, epsilon)

    for i in range(d):
        # perturb x0 in dimension i
        x_perturb = x0.copy()
        x_perturb[i] += epsilon

        grad_i = approx_fprime(x_perturb, fun, epsilon)
        # row i of Hessian = (grad_i - grad0)/ epsilon
        H[i,:] = (grad_i - grad0) / epsilon

    # symmetrize
    H = 0.5 * (H + H.T)
    return H

def compute_info_scipy(model, epsilon=1e-6):
    """
    Use SciPy to approximate the Hessian of the negative log-likelihood 
    for the GH mixture, then invert to get the empirical info matrix.

    Returns: I_e, Cov, SE
    """
    # Flatten final parameters
    X = model.data
    n, p = X.shape
    K = len(model.pi_new)

    theta_hat = flatten_mgh_params(model.pi_new, model.alpha_new)
    d = len(theta_hat)

    # define partial function
    def nll_wrapper(th):
        return full_neg_loglike(th, X, K, p)

    # compute Hessian
    H = approx_hessian_scipy(nll_wrapper, theta_hat, epsilon)
    # H is the Hessian of the negative log-likelihood => observed info

    # invert
    try:
        Cov = np.linalg.inv(H)
    except np.linalg.LinAlgError:
        Cov = np.linalg.pinv(H)

    diag_cov = np.diag(Cov)
    diag_cov_clipped = np.where(diag_cov>0, diag_cov, 0)
    SE = np.sqrt(diag_cov_clipped)
    SE[SE == 0.0] = 0.0001

    return H, Cov, SE

def flatten_mgh_params(pi_list, alpha_list):
    """
    Flatten mixture parameters into a 1D NumPy array.

    Parameters
    ----------
    pi_list : array-like, shape (K,)
        Mixture weights for each of the K components.
    alpha_list : list of length K
        Each element is [lambda, chi, psi, mu, Sigma, gamma], where:
          - lambda, chi, psi : float (scalars)
          - mu : shape (p,)
          - Sigma : shape (p,p)
          - gamma : shape (p,)

    Returns
    -------
    theta : 1D np.ndarray
        The flattened parameter vector.
    """
    K = len(pi_list)
    final_list = []

    # 1) Append the mixing weights
    #    shape (K,) => K floats
    final_list.extend(pi_list)

    # 2) For each cluster k, append the GH parameters in a fixed order
    for k in range(K):
        lam_k, chi_k, psi_k, mu_k, Sigma_k, gamma_k = alpha_list[k]

        # scalars
        final_list.append(lam_k)
        final_list.append(chi_k)
        final_list.append(psi_k)

        # mu_k shape (p,)
        final_list.extend(mu_k)

        # Sigma_k shape (p,p) => flatten row-wise
        final_list.extend(Sigma_k.flatten())

        # gamma_k shape (p,)
        final_list.extend(gamma_k)

    # Convert everything to a float array
    theta = np.array(final_list, dtype=float)
    return theta


def unflatten_mgh_params(theta, p, K):
    """
    Inverse of flatten_mgh_params.

    Parameters
    ----------
    theta : 1D np.ndarray
        Flattened parameter vector (see flatten_mgh_params).
    p : int
        Dimension of data (mu, gamma in R^p).
    K : int
        Number of mixture components.

    Returns
    -------
    pi_list : np.ndarray, shape (K,)
        Mixture weights.
    alpha_list : list of length K
        alpha_list[k] = [lambda, chi, psi, mu, Sigma, gamma].
    """
    idx = 0

    # 1) Extract mixing proportions
    pi_list = theta[idx : idx + K]
    idx += K

    alpha_list = []

    # 2) Extract each cluster's GH parameters
    for _ in range(K):
        lam_k = theta[idx]
        chi_k = theta[idx + 1]
        psi_k = theta[idx + 2]
        idx += 3

        mu_k = theta[idx : idx + p]
        idx += p

        Sigma_flat = theta[idx : idx + p * p]
        idx += p * p
        Sigma_k = Sigma_flat.reshape((p, p))

        gamma_k = theta[idx : idx + p]
        idx += p

        alpha_list.append([lam_k, chi_k, psi_k, mu_k, Sigma_k, gamma_k])

    return pi_list, alpha_list




def single_loglike_j(theta, x_j, p, K):
    """
    Return log( sum_{k=1 to K} pi_k * GHpdf_k(x_j) ) 
    given the flattened parameter vector theta.

    x_j: shape (p,)
    p: dimension of data
    K: number of mixture components
    """
    # 1) unflatten
    pi_list, alpha_list = unflatten_mgh_params(theta, p, K)

    # 2) compute mixture pdf = sum_k pi_k * GHpdf_k( x_j; alpha_list[k] )
    #    alpha_list[k] => [lambda, chi, psi, mu, Sigma, gamma]
    mixture_val = 0.0
    for k in range(K):
        pi_k = pi_list[k]
        lmbd, chi, psi, mu, Sigma, gvec = alpha_list[k]
        pdf_k = np.exp( logpdfmgh(x_j[None], lmbd, chi, psi, mu, Sigma, gvec) )[0]
        # logpdf returns shape (1,) => pdf is exp of that
        mixture_val += pi_k * pdf_k

    # 3) log of that sum
    # if mixture_val <= 0 => numeric issues => clip
    return np.log(max(mixture_val, 1e-300))




def numeric_gradient_single_obs(x_j, theta, p, K, eps=1e-6):
    """
    Central-difference gradient approximation for single_loglike_j(...) 
    at point 'theta'.
    """
    d = len(theta)
    grad = np.zeros(d)

    for i in range(d):
        theta_plus = theta.copy()
        theta_minus = theta.copy()
        theta_plus[i]  += eps
        theta_minus[i] -= eps

        f_plus  = single_loglike_j(theta_plus,  x_j, p, K)
        f_minus = single_loglike_j(theta_minus, x_j, p, K)
        grad[i] = (f_plus - f_minus) / (2 * eps)

    return grad




def compute_empirical_info_numeric(model, eps=1e-6):
    X = model.data
    n, p = X.shape
    K = len(model.pi_new)

    pi_list = model.pi_new
    alpha_list = model.alpha_new
    theta_hat = flatten_mgh_params(pi_list, alpha_list)
    d = len(theta_hat)

    I_e = np.zeros((d, d))
    for j in range(n):
        grad_j = numeric_gradient_single_obs(X[j], theta_hat, p, K, eps=eps)
        I_e += np.outer(grad_j, grad_j)

    # Optionally add ridge
    # I_e += 1e-8 * np.eye(d)

    # invert
    try:
        Cov = np.linalg.inv(I_e)
    except np.linalg.LinAlgError:
        Cov = np.linalg.pinv(I_e)
    
    diag_cov = np.diag(Cov)
    diag_cov_clipped = np.where(diag_cov > 0, diag_cov, 0)
    SE = np.sqrt(diag_cov_clipped)

    return I_e, Cov, SE










def extract_cluster_params(mixture_dict):
    """
    Extracts and structures the cluster parameters from the given mixture dictionary.
    
    Parameters:
        mixture_dict (dict): A dictionary containing cluster information and mixing proportions.
    
    Returns:
        list: A list of lists where each sublist corresponds to a cluster and contains:
              [cpl[1], cpl[0], mu, sigma, alpha].
    """
    # Get the number of clusters from the length of 'pi'
    num_clusters = len(mixture_dict.get("pi", []))
    
    # Initialize the result list
    result = []
    
    for i in range(num_clusters):
        if i in mixture_dict:
            cluster = mixture_dict[i]
            # Extract parameters in the required order
            cpl1 = cluster["cpl"][1]
            cpl0 = cluster["cpl"][0]
            mu = cluster["mu"].tolist()
            sigma = cluster["sigma"]
            alpha = cluster["alpha"].tolist()
            chi, psi = _alphabar2chipsi(cpl0, cpl1)
            result.append([cpl1[0], chi,psi, mu, sigma, alpha])
    
    return result






class MixMGH(BaseMixture):
    def __init__(self,n_clusters,tol=0.0001,initialization="kmeans",print_log_likelihood=False,max_iter=25, verbose=True):
        super().__init__(n_clusters=n_clusters, EM_type="Soft", mixture_type="identical", tol=tol, print_log_likelihood=print_log_likelihood, max_iter=max_iter, verbose=verbose)
        self.k=n_clusters
        self.initialization=initialization
        self.family = "mgh"
    def fit(self, sample):
        """
        Fit the mixture of generalized hyperbolic distributions model.
        
        Parameters:
            X (array-like): data matrix with shape (n_samples, n_features)
            gpar0 (dict, optional): initial parameter dictionary for the GH mixture.
            n_iter (int, optional): maximum number of iterations for the EM algorithm.
            label (array-like, optional): known cluster labels for semi-supervised learning.
        
        Returns:
            self: the fitted model.
        """
        start_time = time.time()
        self.data=self._process_data(sample)
        self.n, self.p = self.data.shape
        self.total_parameters = (self.k -1) + self.k* (2*self.p + (self.p*(self.p+1)/2) +3)
        # Call the mainMGHD function from MixGHD.py to perform the EM algorithm.
        self.results = mainMGHD(data=self.data, G=self.n_clusters, n= self.max_iter,
                                 eps=self.tol)
        self.log_likelihoods = self.results.get("loglik", [])
        self.fitted = True
        self.pi_new = self.results["gpar"]["pi"]
        self.alpha_new = extract_cluster_params(self.results["gpar"])
        self.cluster = self.results["map"]
        self.gamma_temp_ar = self.results["z"]
        self.log_likelihood_new = self.log_likelihoods[-1]
        if self.verbose:
            print("Mixtures of Generalized Hyperbolic Fitting Done Successfully")
        end_time = time.time()
        self.execution_time=end_time-start_time

        
    def get_params(self):
        #print("The estimated pi values are ", self.pi_new)
        #print("The estimated alpha values are ", self.alpha_new)

        return self.pi_new, self.alpha_new

    def predict(self):
        return self.cluster

    def predict_new(self, x):
        x=self._process_data(x)
        data_lol = x.tolist()
        cluster, _ = mixture_clusters(self.gamma_temp_ar, data_lol)

        return cluster


    def responsibilities(self):

        return self.gamma_temp_ar

    def n_iter(self):
        return len(self.log_likelihoods)
    
    def get_info_mat(self, eps=1e-5):
        """
        Compute numeric empirical info matrix & SEs for the final fitted mixture.
        """
        print("Warning: This is experimental; Using Numerical Differentiation")
        # I_e, Cov, SE = compute_empirical_info_numeric(self, eps=eps)
        I_e, Cov, SE = compute_info_scipy(self, epsilon=eps)
        
        return I_e, SE



import numpy as np
from fmvmm.mixtures import skewslashmix_smsn
from fmvmm.mixsmsn.gen import gen_SS_multi
from fmvmm.mixtures.skewslashmix_smsn import dmvSS, d_mixedmvSS
from fmvmm.mixsmsn.information_matrix_smsn import info_matrix_skewslash

def logpdf(x,mu, sigma, lmbda, nu):
    
    return np.log(skewslashmix_smsn.dmvSS(x,mu, sigma, lmbda, nu))

def pdf(x,mu, sigma, lmbda, nu):
    
    return skewslashmix_smsn.dmvSS(x,mu, sigma, lmbda, nu)

def loglike(x,mu, sigma, lmbda, nu):
    
    return np.sum(logpdf(x, mu, sigma, lmbda, nu))

def total_params(mu, sigma, lmbda, nu):
    p = len(mu)
    
    return 2*p + (p*(p+1)/2) +1

def rvs(mu, sigma, lmbda, nu, size = 1):
    
    return gen_SS_multi(size, mu, sigma, lmbda, nu)

def fit(x):
    model = skewslashmix_smsn.SkewSlashMix(1, verbose = False)
    model.fit(x)
    _, alphas = model.get_params()
    
    return alphas[0][0], alphas[0][1], alphas[0][2], np.array([alphas[0][3]])

def info_mat(X,mu,sigma,lmbda,nu):
    
    p =len(mu)
    g = 1
    if isinstance(nu, np.ndarray):
        nu = nu[0]
    IM = info_matrix_skewslash(
    X, [1], [mu], [sigma], [lmbda], nu,
    d_mixedmvSS_func=d_mixedmvSS,
    dmvSS_func=dmvSS,
    g=g, p=p
)
    
    final_IM = expand_reduced_IM_to_full_no_pi_skewslash(IM, p, g)
    
    return final_IM


def expand_reduced_IM_to_full_no_pi_skewslash(IM_reduced, p, g):
    """
    For a Skew-Slash mixture, each cluster j has 'reduced' parameters:
      - mu_j (p)
      - shape_j (p)
      - Sigma_j (p*(p+1)//2)
    => cluster_params_reduced = 2*p + p*(p+1)//2

    Then, if g>1 => (g-1) mixture weights pi_j,
    plus 1 param for nu.

    => total dimension of IM_reduced:
         M_reduced_total = g*(2p + p*(p+1)//2) + (g-1) + 1

    This function:
      1) Removes the pi_j block if g>1 (the (g-1) rows/cols).
      2) Expands each Sigma_j from p*(p+1)//2 to p^2 by duplicating symmetric entries.
      3) Leaves mu_j, shape_j, and the final param nu in place.

    The final dimension is:
      M_full_no_pi = g*(2p + p^2) + 1

    Returns
    -------
    IM_full_no_pi : (M_full_no_pi, M_full_no_pi) np.ndarray
        The expanded info matrix with no pi_j parameters
        and full (p^2) blocks for each Sigma_j.
    """
    cluster_params_reduced = 2*p + (p*(p+1)//2)
    # total dimension in the original "reduced" matrix
    M_reduced_total = g*cluster_params_reduced + (g-1 if g>1 else 0) + 1

    if IM_reduced.shape != (M_reduced_total, M_reduced_total):
        raise ValueError(
            f"Expected IM_reduced to be {M_reduced_total} x {M_reduced_total}, "
            f"but got {IM_reduced.shape}."
        )

    # After removing pi => dimension is
    M_reduced_no_pi = g*cluster_params_reduced + 1

    # After expanding Sigma_j => dimension is
    M_full_no_pi = g*(2*p + p*p) + 1

    # 1) Remove pi block if g>1
    if g > 1:
        offset_pi = g*cluster_params_reduced
        keep_mask = np.ones(M_reduced_total, dtype=bool)
        # pi_j are offset_pi : offset_pi + (g-1)
        pi_indices = np.arange(offset_pi, offset_pi + (g-1))
        keep_mask[pi_indices] = False
        IM_reduced_no_pi = IM_reduced[keep_mask][:, keep_mask]
    else:
        IM_reduced_no_pi = IM_reduced

    if IM_reduced_no_pi.shape != (M_reduced_no_pi, M_reduced_no_pi):
        raise ValueError(
            f"After removing pi, expected shape ({M_reduced_no_pi}, {M_reduced_no_pi}), "
            f"got {IM_reduced_no_pi.shape}"
        )

    # 2) Build duplication matrix D_no_pi => shape (M_full_no_pi, M_reduced_no_pi)
    D_no_pi = np.zeros((M_full_no_pi, M_reduced_no_pi), dtype=float)

    offset_reduced = 0
    offset_full    = 0

    def copy_block_1to1(n_params):
        nonlocal offset_reduced, offset_full
        for k_ in range(n_params):
            D_no_pi[offset_full + k_, offset_reduced + k_] = 1.0
        offset_reduced += n_params
        offset_full    += n_params

    # For each cluster: mu_j(p), shape_j(p), Sigma_j(p*(p+1)//2)
    for j_ in range(g):
        # mu_j => p
        copy_block_1to1(p)
        # shape_j => p
        copy_block_1to1(p)

        # expand Sigma => from p*(p+1)//2 to p^2
        old_start = offset_reduced
        new_start = offset_full
        old_index = 0

        for row_ in range(p):
            for col_ in range(row_, p):
                old_idx = old_start + old_index
                old_index += 1

                new_idx1 = new_start + (row_*p + col_)
                new_idx2 = new_start + (col_*p + row_)

                D_no_pi[new_idx1, old_idx] = 1.0
                if new_idx2 != new_idx1:
                    D_no_pi[new_idx2, old_idx] = 1.0

        offset_reduced += (p*(p+1)//2)
        offset_full    += (p*p)

    # Finally, 1 param for nu
    copy_block_1to1(1)

    IM_full_no_pi = D_no_pi @ IM_reduced_no_pi @ D_no_pi.T
    return IM_full_no_pi



"""
Most of the codes are used from git repository mvem. Huge thanks to the
devoloper. 

"""

import numpy as np
import scipy.stats
from scipy.special import digamma, gammaln
import scipy
from fmvmm.mixsmsn.dens import dmvt_ls, d_mixedmvST
from fmvmm.mixsmsn.information_matrix_smsn import info_matrix_t

def pdf(x, loc, shape, df, allow_singular=True):
    """
    Probability density function of the multivariate Student's t-distribution.
    We use the location-scale parameterisation.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param loc: The location parameter with shape (p,).
    :type loc: np.ndarray
    :param shape: The shape parameter. A positive semi-definite array with
        shape (p, p).
    :type shape: np.ndarray
    :param df: The degrees of freedom of the distribution, > 0.
    :type df: float
    :param allow_singular: Whether to allow a singular matrix.
    :type allow_singular: bool, optional.
    :return: The density at each observation.
    :rtype: np.ndarray with shape (p,).
    """
    return scipy.stats.multivariate_t.pdf(x, loc, shape, df, allow_singular)

def logpdf(x, loc, shape, df,):
    """
    Log-probability density function of the multivariate Student's t-distribution.
    We use the location-scale parameterisation.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param loc: The location parameter with shape (p,).
    :type loc: np.ndarray
    :param shape: The shape parameter. A positive semi-definite array with
        shape (p, p).
    :type shape: np.ndarray
    :param df: The degrees of freedom of the distribution, > 0.
    :type df: float
    :param allow_singular: Whether to allow a singular matrix.
    :type allow_singular: bool, optional.
    :return: The log-density at each observation.
    :rtype: np.ndarray with shape (p,).
    """
    return scipy.stats.multivariate_t.logpdf(x, loc, shape, df)

def loglike(x, loc, shape, df, allow_singular=True):
    """
    Log-likelihood function of the multivariate Student's t-distribution.
    We use the location-scale parameterisation.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param loc: The location parameter with shape (p,).
    :type loc: np.ndarray
    :param shape: The shape parameter. A positive semi-definite array with
        shape (p, p).
    :type shape: np.ndarray
    :param df: The degrees of freedom of the distribution, > 0.
    :type df: float
    :param allow_singular: Whether to allow a singular matrix.
    :type allow_singular: bool, optional.
    :return: The log-likelihood for given all observations and parameters.
    :rtype: float
    """
    return np.sum(logpdf(x, loc, shape, df, allow_singular))

def total_params(loc, shape, df):
    p = len(loc)
    
    return 1 + p + (p*(p+1)/2)

def rvs(loc, shape, df, size=1, random_state=None):
    """
    Random number generator of the multivariate Student's t-distribution.
    We use the location-scale parameterisation.

    :param loc: The location parameter with shape (p,).
    :type loc: np.ndarray
    :param shape: The shape parameter. A positive semi-definite array with
        shape (p, p).
    :type shape: np.ndarray
    :param df: The degrees of freedom of the distribution, > 0.
    :type df: float
    :param size: The number of samples to draw. Defaults to 1.
    :type size: int, optional
    :param random_state: Used for drawing random variates. Defaults to None.
    :type random_state: None, int, np.random.RandomState, np.random.Generator, optional
    :return: The random p-variate numbers generated.
    :rtype: np.ndarray with shape (n, p).
    """
    return scipy.stats.multivariate_t.rvs(loc, shape, df, size, random_state)

def mean(loc, shape, df):
    """
    Mean of the multivariate Student's t-distribution.
    We use the location-scale parameterisation.

    :param loc: The location parameter with shape (p,).
    :type loc: np.ndarray
    :param shape: The shape parameter. A positive semi-definite array with
        shape (p, p).
    :type shape: np.ndarray
    :param df: The degrees of freedom of the distribution, > 0.
    :type df: float
    :return: The mean of the specified distribution.
    :rtype: np.ndarray with shape (p,).
    """
    assert df > 1, "mean of mv student's t is not defined for df <= 1"
    return loc

def var(loc, shape, df):
    """
    Variance of the multivariate Student's t-distribution.
    We use the location-scale parameterisation.

    :param loc: The location parameter with shape (p,).
    :type loc: np.ndarray
    :param shape: The shape parameter. A positive semi-definite array with
        shape (p, p).
    :type shape: np.ndarray
    :param df: The degrees of freedom of the distribution, > 0.
    :type df: float
    :return: The variance of the specified distribution.
    :rtype: np.ndarray with shape (p,).
    """
    assert df > 2, "variance of mv student's t is not defined for df <= 2"
    return (df / (df - 2)) * shape

def fit(X, maxiter = 100, ptol = 1e-6, ftol = 1e-8, return_loglike = False):
    """
    Fit the parameters of the multivariate Student's t-distribution to data
    using an EM algorithm. We use the location-scale parameterisation.

    :param X: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type X: np.ndarray
    :param maxiter: The maximum number of iterations to use in the EM algorithm.
        Defaults to 100.
    :type nit: int, optional
    :param ptol: The relative convergence criterion for the estimated
        parameters. Defaults to 1e-6.
    :type ptol: float, optional
    :param ftol: The relative convergence criterion for the log-likelihood
        function. Defaults to np.inf.
    :type ftol: float, optional
    :param return_loglike: Return a list of log-likelihood values at each iteration.
        Defaults to False.
    :type return_loglike: np.ndarray, optional
    :return: The fitted parameters (<array> mu, <array> scale, <float> df). Also returns
        a list of log-likelihood values at each iteration of the EM algorithm if
        ``return_loglike=True``.
    :rtype: tuple
    """

    # EM Student's t:

    N, D = X.shape
    assert N > D, "Must have more observations than dimensions in _x_"

    # initialise values:
    nu = 4
    mu = np.mean(X, axis = 0)
    sigma = np.linalg.inv((nu / (nu - 2)) * np.cov(X.T))
    log_likelihood = np.array([np.sum(scipy.stats.multivariate_t.logpdf(X, mu, np.linalg.inv(sigma), nu))])

    for i in range(maxiter):

        # save old values
        nu_old = nu
        mu_old = mu
        sigma_old = sigma

        # Expectation Step
        X_ = X - mu
        delta2 = np.sum((X_ @ sigma) * X_, axis=1)
        E_tau = (nu + D) / (nu + delta2)
        E_logtau = digamma(0.5 * (nu + D)) - np.log(0.5 * (nu + delta2))

        # Maximization Step
        mu = (E_tau @ X) / np.sum(E_tau)
        alpha = X - mu
        sigma = np.linalg.inv((alpha * E_tau[:, np.newaxis]).T @ alpha / N)

        # ... if method = "EM"
        func = lambda x: digamma(x/2.) - np.log(x/2.) - 1 - np.mean(E_logtau) + np.mean(E_tau)
        nu = scipy.optimize.fsolve(func, nu) # set min, max ???

        # check for converagence (stopping criterion)
        have_params_converged = \
            np.all(np.abs(mu - mu_old) <= .5 * ptol * (abs(mu) + abs(mu_old))) & \
            np.all(np.abs(nu - nu_old) <= .5 * ptol * (abs(nu) + abs(nu_old))) & \
            np.all(np.abs(sigma - sigma_old) <= .5 * ptol * (abs(sigma) + abs(sigma_old)))
        if ftol < np.inf:
            ll = np.sum(scipy.stats.multivariate_t.logpdf(X, mu, np.linalg.inv(sigma), nu))
            has_fun_converged = np.abs(ll - log_likelihood[-1]) <= .5 * ftol * (
                np.abs(ll) + np.abs(log_likelihood[-1]))
            log_likelihood = np.append(log_likelihood, ll)
        else:
            has_fun_converged = True
        if have_params_converged and has_fun_converged: break

    sigma = np.linalg.inv(sigma)

    if return_loglike:
        return mu, sigma, nu, log_likelihood

    return mu, sigma, nu

def info_mat(X,mu,sigma,nu):
    
    p =len(mu)
    g = 1
    if isinstance(nu, np.ndarray):
        nu = nu[0]
    IM = info_matrix_t(
    X, [1], [mu], [sigma], nu,
    d_mixedmvST_func=d_mixedmvST,
    dmvt_ls_func=dmvt_ls,   
    g=g, p=p)
    
    final_IM = expand_reduced_IM_to_full_no_pi(IM, p, g)
    
    return final_IM

def expand_reduced_IM_to_full_no_pi(IM_reduced, p, g):
    """
    For a Student-t mixture with:
      - cluster_params_reduced = p + (p*(p+1)//2)
          (that is, each cluster has 'mu_j' of length p,
           and Sigma_j of length p*(p+1)//2 for upper-triangular)
      - total dimension = g * cluster_params_reduced + (g-1 if g>1 else 0) + 1
          [the (g-1) for mixture weights if g>1, plus 1 for nu]

    This function:
      1) Removes the rows/columns corresponding to the pi_j (mixture weights),
         if g>1 (if g=1, there are none).
      2) Expands each Sigma_j block to full p^2 by duplicating symmetric entries.
      3) Returns the expanded information matrix, containing mu_j, full Sigma_j,
         and nu, but no pi_j parameters.

    The final dimension is:
      g * (p + p^2) + 1
        => for each cluster: p for mu_j, p^2 for Sigma_j, plus 1 for nu.
    """
    # --- 1) Check dimension of the reduced IM
    cluster_params_reduced = p + (p*(p+1)//2)
    M_reduced_total = g*cluster_params_reduced + (g-1) + 1  # (g-1) mixture weights, 1 for nu
    if IM_reduced.shape != (M_reduced_total, M_reduced_total):
        raise ValueError(
            f"Expected IM_reduced to be {M_reduced_total}x{M_reduced_total}, "
            f"got {IM_reduced.shape}."
        )

    # Dimension after removing pi (if g>1):
    M_reduced_no_pi = g*cluster_params_reduced + 1  # no (g-1)
    # Then the final dimension after expanding Sigma_j to p^2:
    M_full_no_pi = g*(p + p*p) + 1  # each cluster: p + (p^2), plus 1 for nu

    # --- 2) Remove mixture-weight rows/cols (if g>1)
    offset_pi = g*cluster_params_reduced
    if g > 1:
        keep_mask = np.ones(M_reduced_total, dtype=bool)
        # The pi parameters occupy the block [offset_pi : offset_pi + (g-1)]
        pi_indices = np.arange(offset_pi, offset_pi+(g-1))
        keep_mask[pi_indices] = False
        IM_reduced_no_pi = IM_reduced[keep_mask][:, keep_mask]
    else:
        # If g=1, no pi to remove
        IM_reduced_no_pi = IM_reduced

    if IM_reduced_no_pi.shape != (M_reduced_no_pi, M_reduced_no_pi):
        raise ValueError(
            "After removing pi, expected shape "
            f"{(M_reduced_no_pi, M_reduced_no_pi)} but got {IM_reduced_no_pi.shape}"
        )

    # --- 3) Build duplication matrix (D_no_pi) of shape (M_full_no_pi x M_reduced_no_pi)
    # We map from "reduced" param indexing to "full" param indexing.
    D_no_pi = np.zeros((M_full_no_pi, M_reduced_no_pi), dtype=float)

    # We copy mu_j directly (1-to-1) => p parameters,
    # Sigma_j => expand from p(p+1)//2 to p^2.
    offset_reduced = 0
    offset_full    = 0

    def copy_block_1to1(n_params):
        nonlocal offset_reduced, offset_full
        for k_ in range(n_params):
            D_no_pi[offset_full + k_, offset_reduced + k_] = 1.0
        offset_reduced += n_params
        offset_full    += n_params

    # For each cluster j:
    for j_ in range(g):
        # 3.1) mu_j => length p
        copy_block_1to1(p)

        # 3.2) Sigma_j => from p*(p+1)//2 to p^2
        old_start = offset_reduced
        new_start = offset_full
        old_index = 0

        for r_ in range(p):
            for c_ in range(r_, p):
                old_idx = old_start + old_index
                old_index += 1

                new_idx1 = new_start + (r_*p + c_)
                new_idx2 = new_start + (c_*p + r_)

                D_no_pi[new_idx1, old_idx] = 1.0
                if new_idx2 != new_idx1:
                    D_no_pi[new_idx2, old_idx] = 1.0

        offset_reduced += (p*(p+1)//2)
        offset_full    += (p*p)

    # 3.3) We removed pi. So no block for them here.

    # 3.4) final param = nu => 1 param
    copy_block_1to1(1)

    # --- 4) Multiply to get final result:  IM_full_no_pi = D_no_pi * IM_reduced_no_pi * D_no_pi^T
    IM_full_no_pi = D_no_pi @ IM_reduced_no_pi @ D_no_pi.T
    return IM_full_no_pi


"""
Source codes are from R package mixsmsn. 
Converted to python as closely as possible. Many thanks to the authors.
"""

import numpy as np
from scipy.stats import norm, gamma, multivariate_normal
from scipy.linalg import sqrtm

# ----------------------------------------------------------------
#        Auxiliary functions to replicate some R functionality
# ----------------------------------------------------------------

def find_interval(u, intervals):
    """
    Mimic R's findInterval(u, vec). 
    'intervals' should be a non-decreasing sequence, and 'u' is a scalar.
    This returns the index i such that intervals[i] <= u < intervals[i+1].
    """
    # np.searchsorted(intervals, u, side='right') gives the insertion point 
    # to the right. Subtract 1 to get a 0-based index.
    return np.searchsorted(intervals, u, side='right') - 1

def runif(n):
    """Replicate R's runif(n) -> uniform(0,1). Returns an array of length n."""
    return np.random.rand(n)

def rnorm(n, mean=0.0, sd=1.0):
    """Replicate R's rnorm(n, mean, sd)."""
    return np.random.normal(loc=mean, scale=sd, size=n)

def rgamma_r(n, shape, rate):
    """
    Replicate R's rgamma(n, shape, rate).
    In Python/scipy, Gamma is parameterized by shape (alpha) and scale (theta).
    R uses shape (alpha) and rate (beta). scale = 1/beta.
    """
    scale = 1.0 / rate
    return np.random.gamma(shape=shape, scale=scale, size=n)

# ----------------------------------------------------------------
#         Univariate SNI and mixture generation functions
# ----------------------------------------------------------------

def gen_Skew_normal(n, mu, sigma2, shape, nu=None):
    """
    Generate n random values from a Skew-Normal distribution.

    Parameters:
    -----------
    n : int
        Number of samples
    mu : float
        Location parameter
    sigma2 : float
        Scale^2 parameter
    shape : float
        Skewness parameter
    nu : ignored (for compatibility)
    """
    delta = shape / np.sqrt(1.0 + shape**2)
    # abs(rnorm(n)) + sqrt(1 - delta^2)*rnorm(n)
    z1 = np.abs(rnorm(n))
    z2 = rnorm(n)
    out = mu + np.sqrt(sigma2) * (delta * z1 + np.sqrt(1 - delta**2) * z2)
    return out

def gen_Skew_t(n, mu, sigma2, shape, nu):
    """
    Generate n random values from a Skew-t distribution.

    Parameters:
    -----------
    n : int
        Number of samples
    mu : float
        Location parameter
    sigma2 : float
        Scale^2 parameter
    shape : float
        Skewness parameter
    nu : float
        Degrees of freedom
    """
    # (rgamma(n, nu/2, nu/2))^(-1/2) in R means shape = nu/2, rate=nu/2
    # which is scale=2/nu in Python
    gam = rgamma_r(n, nu/2.0, nu/2.0)**(-0.5)  
    # Then multiply by a Skew Normal(0, sigma2, shape):
    sn = gen_Skew_normal(n, 0.0, sigma2, shape)
    return mu + gam * sn

def gen_Skew_cn(n, mu, sigma2, shape, nu1, nu2):
    """
    Generate n random values from a Skew Normal Contaminated (Skew-CN).

    This internally calls rmix(...) with mixture = nu1, 1-nu1.
    That is, with scaled variance sigma2/nu2 in the 'contaminated' component.
    """
    # We generate from the mixture:
    # rmix(n, [nu1, 1-nu1], "Skew.normal", list([mu, sigma2/nu2, shape], [mu, sigma2, shape]))
    out = rmix(n, [nu1, 1 - nu1],
               "Skew.normal",
               [[mu, sigma2/nu2, shape], [mu, sigma2, shape]],
               cluster=False)
    return out

def gen_Skew_slash(n, mu, sigma2, shape, nu):
    """
    Generate n random values from a Skew-Slash distribution.

    We use the fact that if U ~ Uniform(0,1), then U^(1/nu) inverts 
    the distribution for the slash's mixing variable.
    """
    u = runif(n)
    # u2 = u^(1/nu)
    u2 = u**(1.0/nu)
    # Then multiply a skew-normal(0, sigma2, shape) by (u2)^(-1/2).
    sn = gen_Skew_normal(n, 0.0, sigma2, shape)
    return mu + (u2**(-0.5))*sn

def rmix(n, pii, family, arg, cluster=False):
    """
    Generate n samples from a mixture of univariate distributions.

    Parameters:
    -----------
    n : int
        Number of samples
    pii : list or array
        Mixing proportions (length g)
    family : str
        One of ["Normal", "Skew.normal", "t", "Skew.t", "Skew.cn", "Skew.slash"].
    arg : list of lists
        Each element corresponds to the parameters for the chosen family.
        E.g., for a 2-component mixture of Skew.normal, 
        arg might be [[mu1, sigma2_1, shape1], [mu2, sigma2_2, shape2]]
    cluster : bool
        If True, returns also the cluster labels.

    Returns:
    --------
    x : 1D numpy array of length n
    or
    (x, clusters) if cluster=True
    """

    # Validate family
    valid_families = ["t", "Skew.t", "Skew.cn", "Skew.slash", "Skew.normal", "Normal"]
    if family not in valid_families:
        raise ValueError(f"Family {family} not recognized.")

    # Decide which generator to use
    if family in ["Normal", "Skew.normal"]:
        rF1 = gen_Skew_normal
        # If "Normal", forcibly set shape=0 for each arg
        if family == "Normal":
            for a in arg:
                # a is [mu, sigma2, shape]
                a[2] = 0.0

    elif family in ["t", "Skew.t"]:
        rF1 = gen_Skew_t
        # If "t", forcibly set shape=0 
        if family == "t":
            for a in arg:
                a[2] = 0.0

    elif family == "Skew.cn":
        rF1 = gen_Skew_cn

    elif family == "Skew.slash":
        rF1 = gen_Skew_slash

    # We'll generate intervals for mixture selection
    g = len(pii)
    interval = np.zeros(g+1)
    for j in range(1, g):
        interval[j] = interval[j-1] + pii[j-1]
    interval[g] = 1.0

    x1 = np.empty(n, dtype=float)
    clusters = np.empty(n, dtype=int)

    for i in range(n):
        u = np.random.rand()
        comp = find_interval(u, interval)
        clusters[i] = comp
        # param set = arg[comp]
        # E.g. if rF1 = gen_Skew_normal, we call 
        # gen_Skew_normal(1, mu, sigma2, shape, [nu if needed])
        # Because we said each arg[i] might be [mu, sigma2, shape, (nu...)].
        x1[i] = single_call_univariate(rF1, arg[comp])

    if cluster:
        return x1, clusters
    else:
        return x1

def single_call_univariate(fun, param_list):
    """
    Helper to call the univariate generator with a variable number of parameters.
    Param list might be 3, 4, or 5 items depending on the family.
    For example, for Skew.normal we expect [mu, sigma2, shape] (3 args).
    For Skew.t we expect [mu, sigma2, shape, nu] (4 args).
    etc.
    """
    return fun(1, *param_list)[0] if len(param_list) >= 3 else fun(1, *param_list)[0]


# ----------------------------------------------------------------
#        Multivariate SNI and mixture generation functions
# ----------------------------------------------------------------

def gen_SN_multi(n, mu, Sigma, shape, nu=None):
    """
    Generate n random values from a (multivariate) Skew-Normal distribution.

    mu : 1D array (length p)
    Sigma : (p x p) covariance matrix
    shape : 1D array (length p)
    nu : unused, included for API symmetry
    """
    mu = np.asarray(mu)
    shape = np.asarray(shape)
    p = mu.shape[0]
    # delta is a vector
    denom = np.sqrt(1.0 + np.dot(shape, shape))
    delta = shape / denom  # p-dimensional
    # sqrt(Sigma) using the symmetric square root
    sqrt_Sigma = sqrtm(Sigma)

    samples = np.zeros((n, p))
    I_p = np.eye(p)
    # A = sqrt(I - delta delta^T)
    M = I_p - np.outer(delta, delta)
    sqrt_M = sqrtm(M)  # p x p matrix

    for i in range(n):
        # Generate a single sample
        z1 = np.abs(rnorm(1))[0]   # scalar
        z2 = rnorm(p)              # p-dim standard normal
        # Combine
        vec = delta * z1 + sqrt_M @ z2
        # Multiply by sqrt_Sigma and shift by mu
        samples[i, :] = mu + (sqrt_Sigma @ vec)

    return samples

def gen_ST_multi(n, mu, Sigma, shape, nu):
    """
    Generate n random values from a (multivariate) Skew-t distribution.
    """
    mu = np.asarray(mu)
    p = mu.shape[0]
    # (rgamma(n, nu/2, nu/2))^(-1/2)
    gam = rgamma_r(n, nu/2.0, nu/2.0)**(-0.5)
    # We first generate Skew Normal(0, Sigma, shape)
    sn = gen_SN_multi(n, np.zeros(p), Sigma, shape)
    # Then scale each row by gam[i]
    for i in range(n):
        sn[i, :] *= gam[i]
    return (sn + mu)

def gen_SCN_multi(n, mu, Sigma, shape, nu):
    """
    Generate from the multivariate Skew Normal Contaminated (Skew-CN).
    nu is expected to be something like [nu1, nu2].
    """
    mu = np.asarray(mu)
    p = mu.shape[0]
    out = np.zeros((n, p))
    for i in range(n):
        u = np.random.rand()
        if u < nu[0]:
            # shape remains the same, but the covariance is scaled by 1/nu[1]
            out[i,:] = gen_SN_multi(1, mu, Sigma/nu[1], shape)[0]
        else:
            out[i,:] = gen_SN_multi(1, mu, Sigma, shape)[0]
    return out

def gen_SS_multi(n, mu, Sigma, shape, nu):
    """
    Generate from the multivariate Skew-Slash distribution.
    """
    mu = np.asarray(mu)
    p = mu.shape[0]
    u = runif(n)
    u2 = u**(1.0/nu)   # the mixing variable
    # Generate Skew-Normal(0, Sigma, shape)
    sn = gen_SN_multi(n, np.zeros(p), Sigma, shape)
    for i in range(n):
        sn[i, :] *= (u2[i]**-0.5)
    return (sn + mu)

def rmmix(n, pii, family, arg, cluster=False):
    """
    Generate n samples from a mixture of multivariate distributions.

    Parameters:
    -----------
    n : int
        Number of samples
    pii : list or array
        Mixing proportions (length g)
    family : str
        One of ["Normal","Skew.normal","t","Skew.t","Skew.cn","Skew.slash"].
    arg : list of dicts
        Each dict should have keys: mu, Sigma, shape, (nu)...
        Example (for Skew.normal):
           [
             {'mu': [...], 'Sigma': [...], 'shape': [...]},
             {'mu': [...], 'Sigma': [...], 'shape': [...]},
             ...
           ]
    cluster : bool
        If True, return (X, z) where z are mixture labels, else just X.

    Returns:
    --------
    X : (n x p) numpy array
    or 
    (X, clusters) if cluster=True
    """

    valid_families = ["Normal", "Skew.normal", "t", "Skew.t", "Skew.cn", "Skew.slash"]
    if family not in valid_families:
        raise ValueError(f"Family {family} not recognized.")

    # Decide generator
    if family == "Normal":
        rF1 = gen_SN_multi
        # forcibly set shape=0
        for a in arg:
            a['shape'] = np.zeros_like(a['shape'])

    elif family == "Skew.normal":
        rF1 = gen_SN_multi

    elif family in ["t", "Skew.t"]:
        rF1 = gen_ST_multi
        if family == "t":
            # forcibly set shape=0
            for a in arg:
                a['shape'] = np.zeros_like(a['shape'])

    elif family == "Skew.cn":
        rF1 = gen_SCN_multi

    elif family == "Skew.slash":
        rF1 = gen_SS_multi

    g = len(pii)
    interval = np.zeros(g+1)
    for j in range(1, g):
        interval[j] = interval[j-1] + pii[j-1]
    interval[g] = 1.0

    # dimension of the data
    p = len(arg[0]['mu']) 
    X = np.empty((n, p), dtype=float)
    clusters = np.empty(n, dtype=int)

    for i in range(n):
        u = np.random.rand()
        comp = find_interval(u, interval)
        clusters[i] = comp
        # Now we call rF1(1, **arg[comp])
        # for convenience, let's do a helper
        X[i, :] = single_call_multivariate(rF1, arg[comp])

    if cluster:
        return X, clusters
    else:
        return X

def single_call_multivariate(fun, param_dict):
    """
    Helper to call the multivariate generator with arguments in a dictionary.
    We expect something like:
        fun(1, mu, Sigma, shape, nu)
    or in general up to 4 or 5 parameters. We'll handle them carefully.
    """
    # We check keys: mu, Sigma, shape, maybe nu, maybe we ignore the rest
    if 'nu' in param_dict:
        samples = fun(1, param_dict['mu'], param_dict['Sigma'], param_dict['shape'], param_dict['nu'])
    else:
        samples = fun(1, param_dict['mu'], param_dict['Sigma'], param_dict['shape'])
    return samples[0, :]  # single sample row

